import os

SEAFILE_URL = "https://seafile.unistra.fr"
REPO_ID = "c6458c20-0251-4c07-b296-e72240bd7317"
REPO_TOKEN = "6d6011eac0fe2f9e916d2c6a88f7e5fadad4ab10"
ARCHIVE_NAME = "data.zip"
UNISTRA_USERNAME = os.environ.get("UNISTRA_USERNAME")
UNISTRA_PASSWORD = os.environ.get("UNISTRA_PASSWORD")
